#include <stdlib.h>
#include <string.h>

#include "internal.h"


jfauderr pcmbuffer_allocate(PcmBuffer *buf, unsigned int numsamples,
		unsigned int samplerate, unsigned char channels, unsigned char bitspersample)
{
	if (!buf) return jfauderr_genericerror;

	pcmbuffer_zero(buf);

	buf->data = calloc(numsamples, (size_t)channels * bitspersample / 8);
	if (!buf->data) return jfauderr_outofmemory;

	buf->samplerate = samplerate;
	buf->channels = channels;
	buf->bitspersample = bitspersample;
	buf->numsamples = numsamples;

	return jfauderr_noerror;
}

jfauderr pcmbuffer_zero(PcmBuffer *buf)
{
	if (!buf) return jfauderr_genericerror;

	memset(buf,0,sizeof(PcmBuffer));

	return jfauderr_noerror;
}

jfauderr pcmbuffer_free(PcmBuffer *buf)
{
	if (!buf) return jfauderr_genericerror;

	if (buf->data) free(buf->data);
	pcmbuffer_zero(buf);
	
	return jfauderr_noerror;
}

jfauderr pcmbuffer_freemem(PcmBuffer *buf)
{
	if (!buf) return jfauderr_genericerror;

	if (buf->data) free(buf->data);
	buf->data = NULL;
	
	return jfauderr_noerror;
}

jfauderr midibuffer_allocate(MidiBuffer *buf, unsigned int numtracks)
{
	if (!buf) return jfauderr_genericerror;

	midibuffer_zero(buf);

	buf->tracklength = (unsigned int *)calloc(numtracks, sizeof(unsigned int));
	if (!buf->tracklength) return jfauderr_outofmemory;

	buf->tracks = (unsigned char **)calloc(numtracks, sizeof(unsigned char *));
	if (!buf->tracks) {
		free(buf->tracklength);
		return jfauderr_outofmemory;
	}

	buf->numtracks = numtracks;

	return jfauderr_noerror;
}

jfauderr midibuffer_allocatetrack(MidiBuffer *buf, unsigned int track, unsigned int length)
{
	if (!buf) return jfauderr_genericerror;
	if (track >= buf->numtracks) return jfauderr_genericerror;

	buf->tracks[track] = (unsigned char *)calloc(1,length);
	if (!buf->tracks[track]) return jfauderr_outofmemory;

	buf->tracklength[track] = length;

	return jfauderr_noerror;
}

jfauderr midibuffer_zero(MidiBuffer *buf)
{
	if (!buf) return jfauderr_genericerror;

	memset(buf, 0, sizeof(MidiBuffer));

	return jfauderr_noerror;
}

jfauderr midibuffer_free(MidiBuffer *buf)
{
	if (!buf) return jfauderr_genericerror;

	if (buf->tracklength) free(buf->tracklength);
	if (buf->tracks) {
		unsigned i;
		
		for (i=0;i<buf->numtracks;i++)
			if (buf->tracks[i])
				free(buf->tracks[i]);
		free(buf->tracks);
	}
	midibuffer_zero(buf);

	return jfauderr_noerror;
}


#define STALEAGE 256
static BufferLink *managed_head = NULL, *managed_tail = NULL;

jfauderr buffer_init(void)
{
	managed_head = managed_tail = NULL;
	return jfauderr_ok;
}

jfauderr buffer_uninit(void)
{
	BufferLink *lp;

	while (managed_head) {
		lp = managed_head->next;

		switch (managed_head->buffer.type) {
			case buffertype_pcm:
				openal_releasebuffer(&managed_head->buffer.buf.pcm);
				pcmbuffer_free(&managed_head->buffer.buf.pcm);
				break;

			case buffertype_midi:
				midibuffer_free(&managed_head->buffer.buf.midi);
				break;
		}
		
		free(managed_head);
		managed_head = lp;
	}
	managed_head = managed_tail = NULL;
	return jfauderr_ok;
}

jfauderr buffer_fetch(fileident ident, Buffer **buf)
{
	BufferLink *lp;

	lp = managed_head;
	while (lp) {
		if (lp->ident.ident == ident.ident) {
			*buf = &lp->buffer;
			lp->ref++;
			debuglog(DEBUGLVL_GENERAL,jsprintf("buffer: Fetched ident=" fileidentfmt, fileidentval(ident.)));
			return jfauderr_ok;
		}
		lp = lp->next;
	}

	*buf = NULL;
	return jfauderr_ok;
}

jfauderr buffer_isavailable(fileident ident, int *type)
{
	BufferLink *lp;

	lp = managed_head;
	while (lp) {
		if (lp->ident.ident == ident.ident) {
			if (type) *type = lp->buffer.type;
			return jfauderr_ok;
		}
		lp = lp->next;
	}

	return jfauderr_notloaded;
}

jfauderr buffer_release(fileident ident)
{
	BufferLink *lp;

	lp = managed_head;
	while (lp) {
		if (lp->ident.ident == ident.ident) {
			if (lp->ref > 0) {
				lp->ref--;
				lp->age = 0;
				debuglog(DEBUGLVL_GENERAL,jsprintf("buffer: Released ident=" fileidentfmt, fileidentval(ident.)));
			}

			return jfauderr_ok;
		}
		lp = lp->next;
	}

	return jfauderr_notloaded;
}

jfauderr buffer_manage(fileident ident, Buffer *buf)
{
	// allocate an entry on the management list
	BufferLink *lp;

	lp = managed_head;
	while (lp) {
		if (lp->ident.ident == ident.ident) break;
		lp = lp->next;
	}

	if (lp) {
		// replace an entry
		switch (managed_head->buffer.type) {
			case buffertype_pcm:
				openal_releasebuffer(&lp->buffer.buf.pcm);
				pcmbuffer_free(&lp->buffer.buf.pcm);
				break;

			case buffertype_midi:
				midibuffer_free(&lp->buffer.buf.midi);
				break;
		}
	} else {
		lp = (BufferLink*)calloc(1, sizeof(BufferLink));
		if (!lp) return jfauderr_outofmemory;
		
		if (!managed_tail) {
			managed_head = managed_tail = lp;
		} else {
			managed_tail->next = lp;
			lp->prev = managed_tail;
			managed_tail = lp;
		}

		lp->ref = 0;
		lp->ident = ident;
	}

	memcpy(&lp->buffer, buf, sizeof(Buffer));
	debuglog(DEBUGLVL_GENERAL,jsprintf("buffer: Added ident=" fileidentfmt " for management",fileidentval(ident.)));

	return jfauderr_ok;
}

static BufferLink *updateptr = NULL;
jfauderr buffer_update(void)
{
	BufferLink *lpb;

	if (!updateptr) updateptr = managed_head;
	if (!updateptr) return jfauderr_ok;

	if (updateptr->ref > 0) {
		updateptr = updateptr->next;
	} else if (updateptr->age < STALEAGE) {
		updateptr->age++;
	} else {
		lpb = updateptr->next;

		if (updateptr->prev) updateptr->prev->next = updateptr->next;
		else {
			managed_head = updateptr->next;
			if (managed_head) managed_head->prev = NULL;
		}
		if (updateptr->next) updateptr->next->prev = updateptr->prev;
		else {
			managed_tail = updateptr->prev;
			if (managed_tail) managed_tail->next = NULL;
		}
			
		switch (updateptr->buffer.type) {
			case buffertype_pcm:
				openal_releasebuffer(&updateptr->buffer.buf.pcm);
				pcmbuffer_free(&updateptr->buffer.buf.pcm);
				break;

			case buffertype_midi:
				midibuffer_free(&updateptr->buffer.buf.midi);
				break;
		}
		
		debuglog(DEBUGLVL_GENERAL,jsprintf("buffer: Removed stale ident=" fileidentfmt, fileidentval(updateptr->ident.)));
		free(updateptr);

		updateptr = lpb;
	}

	return jfauderr_ok;
}
