#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "internal.h"

typedef struct {
	WaveformFile_defn

	JFAudRawFormat *fmt;
	unsigned long datalen, blocksize, datapos;
} Raw;


// SoundFileReader functions

static jfauderr init(void)
{
	return jfauderr_noerror;
}

static jfauderr uninit(void)
{
	return jfauderr_noerror;
}

static jfauderr test(File *file)
{
	Raw raw;

	if (!file) return jfauderr_genericerror;

	memset(&raw, 0, sizeof(raw));
	
	raw.fmt = (JFAudRawFormat *)file->sf;
	file->sf = NULL;
	
	sysf_seek(&file->fh, 0, SEEK_SET);

	raw.type = type_wave;
	raw.format = fmt_raw;
	
	raw.datalen = sysf_filelen(&file->fh);
	raw.blocksize = raw.fmt->channels * raw.fmt->bitspersample / 8;
	raw.datapos = 0;

	file->sf = (SoundFile *)malloc(sizeof(Raw));
	if (!file->sf) return jfauderr_outofmemory;	// out of memory

	memcpy(file->sf, &raw, sizeof(Raw));

	return jfauderr_noerror;
}

static jfauderr unload(File *file UNUSED, char *noclose)
{
	// caller cleans up file->sf
	*noclose = 0;

	return jfauderr_noerror;
}

static jfauderr getmetainfo(File *file UNUSED, JFAudMetaInfo what UNUSED, char *buf UNUSED, int *len UNUSED)
{
	return jfauderr_notimpl;
}


// WaveformFileReader functions
static jfauderr loadsound(File *file, Raw *raw, PcmBuffer *buf, unsigned long samples, char loop)
{
	long readbytes, justrewound = 0;
	jfauderr err;
	char *rp;

	err = pcmbuffer_allocate(buf, samples, raw->fmt->samplerate, raw->fmt->channels, raw->fmt->bitspersample);
	if (err != jfauderr_noerror) return err;
	buf->numsamples = 0;
	
	rp = (char *)buf->data;
	do {
		readbytes = sysf_read(&file->fh, rp, min(raw->datalen - raw->datapos, samples * raw->blocksize));
		debuglog(DEBUGLVL_TRACE, jsprintf("sysf_read() = %ld",readbytes));
		if (readbytes < 0) {
			// read failures are always fatal
			pcmbuffer_free(buf);
			return jfauderr_ioerror;
		} else if (readbytes == 0) {
			if (!loop) {
				// zero bytes read is fatal only if not looping
				pcmbuffer_free(buf);
				break;
			} else {
				// otherwise we rewind
				if (justrewound) loop = 0;	// after rewinding we read 0, so break the loop
				else {
					sysf_seek(&file->fh, 0, SEEK_SET);
					raw->datapos = 0;
					justrewound = 1;
				}
			}
		} else {
			unsigned long i,s;
			s = readbytes / raw->blocksize;
			buf->numsamples += s;

			samples -= s;
			justrewound = 0;

#if MustSwapBig
			if (raw->fmt->bitspersample == 16 && raw->fmt->bigendian) {
#else
			if (raw->fmt->bitspersample == 16 && !raw->fmt->bigendian) {
#endif
				unsigned short *p;
				// byte swap shorts
				p = (unsigned short *)rp;
				for (i=readbytes/2; i>0; i--, p++) {
					*p = BigShort(*p);
				}
			}
		}

		raw->datapos += readbytes;
		rp += readbytes;

	} while (samples > 0 && loop);

	if (buf->numsamples > 0)
		debuglog(DEBUGLVL_WORDY,jsprintf("Read %d sample(s)", buf->numsamples));

	return jfauderr_noerror;
}

static jfauderr readsamples(File *file, int n, PcmBuffer *buf, char loop)
{
	Raw *raw;

	raw = (Raw *)file->sf;

	if (n <= 0) {
		pcmbuffer_zero(buf);
		return jfauderr_noerror;
	}

	return loadsound(file, raw, buf, n, loop);
}

static jfauderr readallsamples(File *file, PcmBuffer *buf)
{
	Raw *raw;
	unsigned long samples;

	raw = (Raw *)file->sf;

	samples = raw->datalen / raw->blocksize;
	if (samples == 0) {
		pcmbuffer_zero(buf);
		return jfauderr_noerror;
	}

	sysf_seek(&file->fh, 0, SEEK_SET); raw->datapos = 0;
	
	return loadsound(file, raw, buf, samples, 0);
}

static jfauderr getplaytime(File *file, float *len)
{
	Raw *raw;

	raw = (Raw *)file->sf;

	*len = (float)(raw->datalen / raw->blocksize) / (float)raw->fmt->samplerate;
	
	return jfauderr_noerror;
}


WaveformFileReader wf_raw = {
	"RAW PCM",
	init,
	uninit,

	type_wave,
	test,
	unload,
	getmetainfo,

	fmt_raw,
	readsamples,
	readallsamples,
	getplaytime
};

