#include <string.h>
#include <stdlib.h>

#include "internal.h"

#if (WITH_VORBIS == 0)
static jfauderr init(void) { return jfauderr_disabled; }
static jfauderr uninit(void) { return jfauderr_disabled; }
static jfauderr test(File *file UNUSED) { return jfauderr_disabled; }
static jfauderr unload(File *file UNUSED, char *noclose UNUSED) { return jfauderr_disabled; }
static jfauderr getmetainfo(File *file UNUSED, char what UNUSED, char *buf UNUSED, int len UNUSED) { return jfauderr_disabled; }
static jfauderr readsamples(File *file UNUSED, int n UNUSED, PcmBuffer *buf UNUSED, char loop UNUSED) { return jfauderr_disabled; }
static jfauderr readallsamples(File *file UNUSED, PcmBuffer *buf UNUSED) { return jfauderr_disabled; }
static jfauderr getplaytime(File *file UNUSED, float *len UNUSED) { return jfauderr_disabled; }
#else

#ifdef __WATCOMC__
#define __int32 int
#define __int16 short
#endif

#include "vorbis/codec.h"
#include "vorbis/vorbisfile.h"

#if (DYNAMIC_VORBIS == 1)

#ifdef _WIN32
#ifdef PACKED
#undef PACKED
#endif
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#define DLLNAME "vorbisfile.dll"
#else

#include <dlfcn.h>
#define DLLNAME "libvorbisfile.so"
#endif


static struct {
#ifdef _WIN32
	HMODULE dll;
#else
	void *dll;
#endif

	int (*ov_open_callbacks)(void *datasource, OggVorbis_File *vf, char *initial, long ibytes, ov_callbacks callbacks);
	int (*ov_clear)(OggVorbis_File *vf);
	vorbis_info *(*ov_info)(OggVorbis_File *vf,int link);
	long (*ov_seekable)(OggVorbis_File *vf);
	long (*ov_streams)(OggVorbis_File *vf);
	long (*ov_read)(OggVorbis_File *vf,char *buffer,int length, int bigendianp,int word,int sgned,int *bitstream);
	int (*ov_pcm_seek_lap)(OggVorbis_File *vf,ogg_int64_t pos);
	int (*ov_pcm_seek)(OggVorbis_File *vf,ogg_int64_t pos);
	double (*ov_time_total)(OggVorbis_File *vf,int i);
	ogg_int64_t (*ov_pcm_total)(OggVorbis_File *vf,int i);
	vorbis_comment *(*ov_comment)(OggVorbis_File *vf,int i);
} vorbisfile;

#define ov_open_callbacks vorbisfile.ov_open_callbacks
#define ov_clear vorbisfile.ov_clear
#define ov_info vorbisfile.ov_info
#define ov_seekable vorbisfile.ov_seekable
#define ov_streams vorbisfile.ov_streams
#define ov_read vorbisfile.ov_read
#define ov_pcm_seek_lap vorbisfile.ov_pcm_seek_lap
#define ov_pcm_seek vorbisfile.ov_pcm_seek
#define ov_time_total vorbisfile.ov_time_total
#define ov_pcm_total vorbisfile.ov_pcm_total
#define ov_comment vorbisfile.ov_comment

#define testifloaded() { if (vorbisfile.dll == NULL) return jfauderr_notinited; }
#else

#define testifloaded() { }
#endif


static size_t vfread(void *ptr, size_t size, size_t nmemb, void *datasource)
{ return sysf_read(&((File*)datasource)->fh, ptr, size*nmemb); }

static int vfseek(void *datasource, ogg_int64_t offset, int whence)
{ return sysf_seek(&((File*)datasource)->fh, (int)offset, whence); }

static int vfclose(void *datasource UNUSED)
{ return 0; }	// we close ourself

static long vftell(void *datasource)
{ return sysf_tell(&((File*)datasource)->fh); }

static ov_callbacks vfcallbacks = { vfread, vfseek, vfclose, vftell };


typedef struct {
	WaveformFile_defn

	OggVorbis_File vf;

	int channels;
	long samplerate;

	int seekable;	// affects whether we can loop or suck in a whole file
	
	int blocksize;
	ogg_int64_t totalsamples;
} Vorbis;



// SoundFileReader functions

static jfauderr init(void)
{
#if (DYNAMIC_VORBIS == 1)
        int er=0;

#ifdef _WIN32
        void *sym;
        vorbisfile.dll = LoadLibrary(DLLNAME);

#define gpa(d,x) sym = GetProcAddress(d,x); { \
                if (!sym) { \
                        debuglog(DEBUGLVL_GENERAL,jsprintf("Failed getting symbol %s from " DLLNAME,x)); \
                        er = 1; \
                } \
        }
#else
        vorbisfile.dll = dlopen(DLLNAME, RTLD_GLOBAL|RTLD_NOW);
        dlerror();

#define gpa(d,x) dlsym(d,x); { \
                const char *s; \
                if ((s = dlerror())) { \
                        debuglog(DEBUGLVL_GENERAL,jsprintf("Failed getting symbol %s from " DLLNAME,x)); \
                        er = 1; \
                } \
        }
#endif

        if (!vorbisfile.dll) return jfauderr_disabled;
        
        ov_open_callbacks = (void*)gpa(vorbisfile.dll, "ov_open_callbacks");
        ov_clear = (void*)gpa(vorbisfile.dll, "ov_clear");
        ov_info = (void*)gpa(vorbisfile.dll, "ov_info");
        ov_seekable = (void*)gpa(vorbisfile.dll, "ov_seekable");
        ov_streams = (void*)gpa(vorbisfile.dll, "ov_streams");
        ov_read = (void*)gpa(vorbisfile.dll, "ov_read");
        ov_pcm_seek_lap = (void*)gpa(vorbisfile.dll, "ov_pcm_seek_lap");
        ov_pcm_seek = (void*)gpa(vorbisfile.dll, "ov_pcm_seek");
        ov_time_total = (void*)gpa(vorbisfile.dll, "ov_time_total");
        ov_pcm_total = (void*)gpa(vorbisfile.dll, "ov_pcm_total");
	ov_comment = (void*)gpa(vorbisfile.dll, "ov_comment");

        if (er) {
#ifdef _WIN32
                FreeLibrary(vorbisfile.dll);
#else
                dlclose(vorbisfile.dll);
#endif

                vorbisfile.dll = NULL;
                debuglog(DEBUGLVL_GENERAL,"Failed fetching required symbols from " DLLNAME ".");
                return jfauderr_initfailed;
        }
#endif	

	return jfauderr_noerror;
}

static jfauderr uninit(void)
{
#if (DYNAMIC_VORBIS == 1)
        if (vorbisfile.dll) {
#ifdef _WIN32
                FreeLibrary(vorbisfile.dll);
#else
                dlclose(vorbisfile.dll);
#endif

                vorbisfile.dll = NULL;
        }
#endif

	return jfauderr_noerror;
}

static jfauderr test(File *file)
{
        Vorbis vorb;
        int ret;
	vorbis_info *vi;
	long i,j;

        testifloaded();

        sysf_seek(&file->fh, 0, SEEK_SET);
        memset(&vorb,0,sizeof(Vorbis));

        ret = ov_open_callbacks((void*)file, &vorb.vf, NULL, 0, vfcallbacks);
        if (ret == OV_ENOTVORBIS) return jfauderr_notrecognised;
        else if (ret != 0) return jfauderr_genericerror;

	vorb.seekable = (int)ov_seekable(&vorb.vf);

	vi = ov_info(&vorb.vf, 0);
	if (!vi) {
		ov_clear(&vorb.vf);
		return jfauderr_genericerror;
	}

	if (vi->channels != 1 && vi->channels != 2) {
		ov_clear(&vorb.vf);
		return jfauderr_notcompatible;
	}

	vorb.channels = vi->channels;
	vorb.samplerate = vi->rate;
	vorb.blocksize = vorb.channels * 2;
	
	if (vorb.seekable) {
		vorb.totalsamples = ov_pcm_total(&vorb.vf, -1);

		i = ov_streams(&vorb.vf);
		if (i>1) {
			debuglog(DEBUGLVL_GENERAL,jsprintf("%ld chained bitstreams found; checking", i));
			for (j=1;j<i;j++) {
				vi = ov_info(&vorb.vf, j);
				if (!vi) break;
				if (vi->channels != vorb.channels || vi->rate != vorb.samplerate)
					debuglog(DEBUGLVL_GENERAL,jsprintf("  bitstream %ld has different properties to those before it!",j));
			}
		}
	}

	vorb.type = type_wave;
	vorb.format = fmt_vorbis;
	
	file->sf = (SoundFile *)malloc(sizeof(Vorbis));
	if (!file->sf) {
		ov_clear(&vorb.vf);
		return jfauderr_outofmemory;	// out of memory
	}

	memcpy(file->sf, &vorb, sizeof(Vorbis));
	
	return jfauderr_noerror;
}

static jfauderr unload(File *file, char *noclose)
{
	testifloaded();

	ov_clear(&((Vorbis*)file->sf)->vf);
	
	// caller frees file->sf
	*noclose = 0;

	return jfauderr_noerror;
}

static jfauderr getmetainfo(File *file, JFAudMetaInfo what, char *buf, int *len)
{
	vorbis_comment *vc;
	Vorbis *vorb = (Vorbis*)file->sf;
	const char *tagname;
	int i;

	testifloaded();

	vc = ov_comment(&vorb->vf, -1);
	if (!vc) return jfauderr_genericerror;

	switch (what) {
		case JFAudMetaInfo_title: tagname="TITLE="; break;
		case JFAudMetaInfo_artist: tagname="ARTIST="; break;
		default: return jfauderr_genericerror;
	}

	for (i=0;i<vc->comments;i++) {
		if (!strncasecmp(tagname,vc->user_comments[i],strlen(tagname))) {
			*len = min(*len, (int)(vc->comment_lengths[i] - strlen(tagname)));
			memcpy(buf, vc->user_comments[i]+strlen(tagname), *len);
			return jfauderr_noerror;
		}
	}
	*len = 0;

	return jfauderr_noerror;
}


// WaveformFileReader functions

static jfauderr readsamples(File *file, int n, PcmBuffer *buf, char loop)
{
	jfauderr err;
	Vorbis *vorb = (Vorbis *)file->sf;
	char *rp;
	int br, bs, bitsnum, rewound=0;

	testifloaded();
	
	if (n <= 0) {
		pcmbuffer_zero(buf);
		return jfauderr_noerror;
	}

	err = pcmbuffer_allocate(buf, n, vorb->samplerate, vorb->channels, 16);
	if (err != jfauderr_noerror) return err;
	buf->numsamples = 0;

	rp = (char *)buf->data;
	bs = n * vorb->blocksize;
	do {
		br = ov_read(&vorb->vf, rp, bs, 0, 2, 1, &bitsnum);
		if (br == 0 && loop && vorb->seekable) {
			if (rewound) break;
			else rewound = 1;
			br = ov_pcm_seek_lap(&vorb->vf, 0);
			if (br == 0) br = 1;
		} else if (br > 0) {
			bs -= br;
			rp += br;
			rewound = 0;
		}
	} while ((br > 0 || br == OV_HOLE) && bs > 0);

	if (br < 0) {
		pcmbuffer_free(buf);
		return jfauderr_genericerror;
	}
	buf->numsamples = ((long)rp - (long)buf->data) / vorb->blocksize;

	if (buf->numsamples > 0)
		debuglog(DEBUGLVL_WORDY,jsprintf("Read %d sample(s)", buf->numsamples));

	return jfauderr_noerror;
}

static jfauderr readallsamples(File *file, PcmBuffer *buf)
{
	jfauderr err;
	Vorbis *vorb = (Vorbis *)file->sf;
	char *rp;
	int br, bs, bitsnum;

	testifloaded();

	if (!vorb->seekable) {
		pcmbuffer_zero(buf);
		return jfauderr_noerror;
	}
	
	if (vorb->totalsamples <= 0) {
		pcmbuffer_zero(buf);
		return jfauderr_noerror;
	}

	if (ov_pcm_seek(&vorb->vf, 0)) return jfauderr_genericerror;

	err = pcmbuffer_allocate(buf, vorb->totalsamples, vorb->samplerate, vorb->channels, 16);
	if (err != jfauderr_noerror) return err;
	buf->numsamples = 0;
	
	rp = (char *)buf->data;
	bs = vorb->totalsamples * vorb->blocksize;
	do {
		br = ov_read(&vorb->vf, rp, bs, 0, 2, 1, &bitsnum);
		if (br > 0) {
			bs -= br;
			rp += br;
		}
	} while ((br > 0 || br == OV_HOLE) && bs > 0);

	if (br < 0) {
		pcmbuffer_free(buf);
		return jfauderr_genericerror;
	}
	buf->numsamples = ((long)rp - (long)buf->data) / vorb->blocksize;

	debuglog(DEBUGLVL_WORDY,jsprintf("Read %d sample(s)", buf->numsamples));
	
	return jfauderr_noerror;
}

static jfauderr getplaytime(File *file, float *len)
{
	Vorbis *vorb = (Vorbis *)file->sf;
	double l;

	testifloaded();

	l = ov_time_total(&vorb->vf, -1);
	if (l >= 0.0) *len = (float)l;
	else *len = -1.0;
	
	return jfauderr_noerror;
}

#endif // WITH_VORBIS == 0


WaveformFileReader wf_vorbis = {
	"OggVorbis",
	init,
	uninit,

	type_wave,
	test,
	unload,
	getmetainfo,

	fmt_vorbis,
	readsamples,
	readallsamples,
	getplaytime
};

